/* jshint esversion: 6, node: true, unused: true */
'use strict';

const fs      = require('fs');
const http    = require('http');
const dns     = require('dns');
const winston = require('winston');
const rimraf  = require('./rimraf');

function getJSON(connectionString) {
  return new Promise(function(resolve, reject) {
    getText(connectionString).then(function(text) {
      try {
        var data = JSON.parse(text);
        resolve(data);
      } catch (error) {
        reject(error);
      }
    });
  });
}

var getCount = 0;
function getText(connectionString) {
  var id = ++getCount;

  function executor(resolve, reject) {
    winston.debug(`[GET#${ id }] ${ connectionString }`);
    var req = http.get(connectionString, function(response) {
      if (response.statusCode !== 200) return reject(response.statusMessage);
      var text = '';

      response.on('data', function(chunk) { text += chunk; });
      response.on('end', function() {
        winston.debug(`[GET#${ id }] OK`, text);
        resolve(text);
      });
    });
    req.on('error', reject);
  }

  return new Promise(executor).catch(function(error) {
    winston.debug(`[GET#${ id }] ${ error }`);
  });
}

function stat(path) {
  function executor(resolve, reject) {
    fs.stat(path, function(err, stat) {
      if (err) reject(err);
      else resolve(stat);
    });
  }

  return new Promise(executor);
}

function lookup(fqdn) {
  return new Promise(function(resolve, reject) {
    dns.lookup(fqdn, function(error, address, family) {
      if (error) reject(error);
      else resolve(address, family);
    });
  });
}

function move(source, destination) {
  return new Promise(function(resolve, reject) {
    function handleUnlink(e) {
      if (e) reject(e);
      else resolve();
    }
    function handleRename(e) {
      if (!e) return resolve();
      if (e.code !== 'EXDEV') return reject(e);

      copy(source, destination).then(function() {
        fs.unlink(source, handleUnlink);
      }).catch(reject);
    }

    fs.rename(source, destination, handleRename);
  });
}

// asynchrously remove file tree
// when we change the configuration we should remove the old ProtoSync directory
// when we then have a lot of files this'll work much better with how JS actually works
// uses rimraf because they already have solved a lot of problems that could arise
function rmtree(path) {
  return new Promise(function(resolve, reject) {
    stat(path).then(function() {
      rimraf(path, { glob: false }, function(error) {
        if (error) reject(error);
        else resolve();
      });
    }, resolve);
  });
}

function unlink(filepath) {
  return new Promise(function(resolve, reject) {
    fs.unlink(filepath, function(error) {
      if (error) reject(error);
      else resolve();
    });
  });
}

function copy(source, destination) {
  var readStream  = fs.createReadStream(source);
  var writeStream = fs.createWriteStream(destination);

  return new Promise(function(resolve, reject) {
    readStream.on('error', reject);
    writeStream.on('error', reject);
    readStream.on('close', resolve);

    readStream.pipe(writeStream);
  });
}

var readCount = 0;
function read(path, encoding) {
  var id = ++readCount;
  function executor(resolve, reject) {
    winston.debug(`[read#${id}]`, { path: path });
    function callback(err, data) {
      winston.debug(`[read#${id}]`, { result: arguments });
      if (err) reject(err);
      else resolve(data, path);
    }
    fs.readFile(path, encoding || 'utf8', callback);
  }
  return new Promise(executor);
}

var writeCount = 0;
function write(path, data) {
  var id = ++writeCount;
  function executor(resolve, reject) {
    winston.debug(`[write#${id}]`, { path: path });
    function callback(err) {
      winston.debug(`[write#${id}]`, { result: arguments });
      if (err) reject(err);
      else resolve(path);
    }
    fs.writeFile(path, data, callback);
  }
  return new Promise(executor);
}

module.exports = {
  getJSON: getJSON,
  getText: getText,
  lookup:  lookup,
  stat:    stat,
  copy:    copy,
  move:    move,
  unlink:  unlink,
  rmtree:  rmtree,
  read:    read,
  write:   write
};
