/* jshint esversion: 6, node: true, unused: true */
'use strict';

const moduleName = 'native';

const app     = require('app');
const dialog  = require('dialog');

const supportedPlatforms = {
  darwin:  true,
  freebsd: false,
  linux:   true,
  sunos:   false,
  win32:   true
};
var notSupported = supportedPlatforms[process.platform] !== true;

function abort(title, message, code) {
  dialog.showErrorBox(title, message);
  app.exit(code || 1);
}

if (notSupported) {
  var title   = 'Unsupported Platform';
  var message = `${process.platform} is not supported.`;

  abort(title, message);
}

var Native = {};

const winston   = require('winston');

const spawnSync = require('child_process').spawnSync;
const spawn     = require('child_process').spawn;
const exec      = require('child_process').exec;
const path      = require('path');
const helper    = require(path.join(app.getAppPath(), 'helper'));

// If we write into the apps path, Windows Vista, 7, ... will write into a Virtual Store instead (paths will be broken).
//const syncthingHome = app.getPath('userData');
const syncthingHome = path.join(app.getPath('userData'), 'syncthing');
const configPath = app.getPath('userData');

var isWin = /^win/.test(process.platform);

// TODO: these paths are only temporarily set to the application folder
var ssh_key           = path.join(configPath, 'id_rsa');
var ssh_pubkey        = path.join(configPath, 'id_rsa.pub');

if (isWin) {
  var syncthing_client_cmd = path.join(app.getAppPath(), 'syncthing', 'syncthing', 'syncthing.exe');
  var syncthing_cli_cmd    = path.join(app.getAppPath(), 'syncthing', 'syncthing-cli', 'syncthing-cli.exe');
  var proxytunnel_cmd      = `${ app.getAppPath().replace(/\\/g, '/') }/syncthing/proxytunnel/proxytunnel.exe`;
  var ssh_cmd              = path.join(app.getAppPath(), 'syncthing', 'proxytunnel', 'ssh.exe');
  var ssh_keygen_cmd       = path.join(app.getAppPath(), 'syncthing', 'proxytunnel', 'ssh-keygen.exe');

  var shell_cmd = path.join(app.getAppPath(), 'syncthing', 'proxytunnel', 'sh.exe');
  process.env.SHELL = shell_cmd;
} else {
  var syncthing_client_cmd = path.join(app.getAppPath(), 'syncthing', 'syncthing', 'syncthing');
  var syncthing_cli_cmd    = path.join(app.getAppPath(), 'syncthing', 'syncthing-cli', 'syncthing-cli');
  var proxytunnel_cmd      = path.join(app.getAppPath(), 'syncthing', 'proxytunnel', 'proxytunnel');
  // we assume ssh to be installed on Unix systems
  var ssh_cmd              = 'ssh';
  var ssh_keygen_cmd       = 'ssh-keygen';
}

// TODO switch to API calls completely!
// TODO hide syncthing cmd wrappers behind a Syncthing Facade
function syncthing_cli(params) {
    winston.info('[syncthing-cli] Executing...', { params: params });
    var cli = spawnSync(syncthing_cli_cmd, params);
    winston.info('[syncthing-cli] Done.', { params: params, error: cli.error });

    var stdout = cli.stdout.toString();
    if (stdout) winston.debug(`[syncthing-cli] ${stdout}`);

    var stderr = cli.stderr.toString();
    if (stderr) winston.warn(`[syncthing-cli] ${stderr}`);

    return { stderr: stderr, stdout: stdout };
}

// async variant of syncthing_cli, returns promise.
function syncthingCli() {
  var args = Array.prototype.slice.call(arguments);

  return new Promise(function(resolve, reject) {
    winston.info('[syncthing-cli] Executing...', { args: args });

    var cli = spawn(syncthing_cli_cmd, args);
    var std = { out: '', err: '' };

    cli.stderr.on('data', function(buffer) { std.err += buffer.toString(); });
    cli.stdout.on('data', function(buffer) { std.out += buffer.toString(); });

    cli.on('close', function(code) {
      winston.info(`[syncthing-cli] Done.`, { params: args, code: code });
      if (code) reject(code);
      else resolve(std);
    });
  });
}

function syncthingClient() {
  var args = Array.prototype.slice.call(arguments);
  args.push(`-home=${ syncthingHome }`);

  function executor(resolve, reject) {
    spawn(syncthing_cli_cmd, ['status']).on('close', function(code) {
      if (code === 0) {
        winston.info('[syncthing] Already running...');
        return resolve();
      }

      winston.info('[syncthing] Starting...', {
        command: syncthing_client_cmd, arguments: args
      });
      
      // copy environment variables from process
      var productionEnv = Object.create(process.env);
      // set go environment to 1 CPU max - there is no real need for more
      // and this will keep the maximum load for syncthing at 1 
      productionEnv.GOMAXPROCS = '1'; 
      // launch syncthing daemon with GOMAXPROCS = 1 env parameter
      var syncthing = spawn(syncthing_client_cmd, args, {env: productionEnv});

      function resolver(data) {
        var output = data.toString();
        var isListening = /API listening on/.test(output);
        if (isListening) {
          winston.debug('[syncthing] Ready!');
          syncthing.stdout.removeListener('data', resolver);
          resolve();
        }
      }
      syncthing.stdout.on('data', resolver);

      function shutdownSyncthing() { syncthingCli(['shutdown']); }
      app.on('quit', shutdownSyncthing);
      syncthing.on('close', function(error) {
        app.removeListener('quit', shutdownSyncthing);
        if (error) reject(error);
      });

      syncthing.stdout.on('data', logBuffer('debug', '[syncthing]'));
      syncthing.stderr.on('data', logBuffer('warn', '[syncthing]'));
    });
  }

  return new Promise(executor);
}

function add_localhost_to_config(remoteboxid, boxname, callback) {
  var filepath = path.join(syncthingHome, 'config.xml');
  return helper.read(filepath).then(function(data) {
    // NOTE: the DOT does not match newline, so we use [\s\S]+ instead
    // var re = new RegExp("pattern", "flags");

    //TODO: stop naively assuming the first is the correct one!!!
    // RADAR why not doing this via API
    var re = new RegExp("<device id=\"" + remoteboxid + "\" name=[\\s\\S]*?</device>","m");

    winston.debug('[syncthing] Adding localhost to configuration.', {
      remoteBoxId: remoteboxid,
      boxname: boxname
    });
    var newdata = data.replace(re, "<device id=\"" + remoteboxid + "\" name=\"" + boxname + "\" compression=\"metadata\" introducer=\"false\"><address>127.0.0.1:22001</address>\n<address>dynamic</address></device>");
    // <device id="PSNJXZK-2OXHBJV-T5243KI-SHLTEKB-5IKSVOU-SWQSE7K-2LOZE6H-KFBL3Q2" name="PSNJXZK-2OXHBJV-T5243KI-SHLTEKB-5IKSVOU-SWQSE7K-2LOZE6H-KFBL3Q2" compression="metadata" introducer="false">
    //         <address>dynamic</address>
    //             </device>

    return helper.write(filepath, newdata);
  }).then(callback);
}

function getPublicKey() {
  return helper.stat(ssh_pubkey).catch(generateSSHKey).then(function() {
    return helper.read(ssh_pubkey).then(function(key) {
      winston.debug(`[${ moduleName }] Read public key.`, { key: key });
      return Promise.resolve(key);
    });
  });
}

function initSyncthingConfig() {
  // Creates a blank syncthing config, that contains a default folder we dont want
  var command = `"${ syncthing_client_cmd }" -verbose -generate="${ syncthingHome }" -home="${ syncthingHome }"`;

  return new Promise(function(resolve, reject) {
    winston.debug('[syncthing] Generating config...', { command: command });
    var generate = exec(command);

    generate.stderr.on('data', logBuffer('warn', '[syncthing]'));
    generate.stdout.on('data', logBuffer('debug', '[syncthing]'));

    generate.on('close', function (error) {
      if (error) {
        winston.error('[syncthing] Could not generate configuration.');
        reject(error);
      }

      var filepath = path.join(syncthingHome, 'config.xml');
      winston.debug('[syncthing] Configuration generated.', { path: filepath });

      helper.read(filepath).then(function(data) {
        var dataWithoutFolder = data.replace(/<folder[\s\S]+<\/folder>/m, "");
        winston.debug('[syncthing] Removed default folder from configuration.');

        winston.debug('[syncthing] Writing fixed configuration to disk...');
        return helper.write(filepath, dataWithoutFolder);

      }).then(resolve).catch(function(err) {
        winston.error('[syncthing] Configuration failed.');
        reject(err);
      });
    });
  });
}

function logBuffer(level, caption) {
  function logger(buffer) {
    buffer.toString().split("\n").forEach(function(line) {
      winston.log(level, `${ caption } ${ line }`);
    });
  }
  return logger;
}

// TODO implement a KeyManager
// generates
// de-/publishes
// reads
// depublishes generic key
// TODO We should publish the key on generation or we get a lot of f*cks!
function generateSSHKey() {
  var args = [ '-trsa', '-b2048', `-f"${ ssh_key }"`, '-N ""' ];
  winston.info(`[${ moduleName }] Generating SSH key pair...`, {
    command: `${ ssh_keygen_cmd } ${ args.join(' ') }`
  });

  return new Promise(function(resolve, reject) {
    var keygen = exec(`"${ ssh_keygen_cmd }" ${ args.join(' ') }`);

    keygen.stdout.on('data', logBuffer('debug', '[ssh-keygen]'));
    keygen.stderr.on('data', logBuffer('warn', '[ssh-keygen]'));

    keygen.on('close', function(code) {
      if (code) reject(code);
      else resolve();
    });
  });
}

// TODO Implement a TunnelManager
// depends on KeyManager implementation
// opens
// closes
var tunnel, tunnelStarted = false;
function startTunnel(boxname, username) {
    //
    // TODO: ATTENTION!!! boxname currently give full address!!!
    //
  if (tunnelStarted) return Promise.resolve();

  // We have an issue here: if the user is using a custom address for its box
  // then the nginx will likely answer when we try to connect to it via 443
  // to circumvent this FOR NOW the user needs to open its apache on port 80
  // and through that we connect with HTTP CONNECT to the SSH server
  // Since we want to keep to use HTTPS when possible, we need to check if the 
  // supplied address here is a *.protonet.info domain, or not.
  if (/.*\.protonet\.info$/.test(boxname) == true) {
      // supplied boxname is reachable over reverse proxy we use HTTPS
      winston.info("[tunnel] box is reachable over reverse proxy, using port 443");
      var args = [
        `-i"${ ssh_key }"`,
        '-N',
        '-L22001:localhost:22000',
        '-oServerAliveInterval=20',
        '-oStrictHostKeyChecking=no',
        `-oProxyCommand="'${ proxytunnel_cmd }' -E -p ${ boxname }:443 -d protonet:22 -o ${ boxname }"`,
        `${ username }@${ boxname }`
      ];
  } else {
      // supplied boxname is a custom domain (or plain IP) so we use HTTP
      winston.warn("[tunnel] warning! using port 80 for tunnel your connection is encrypted only by ssh but not stealthed!");
      var args = [
        `-i"${ ssh_key }"`,
        '-N',
        '-L22001:localhost:22000',
        '-oServerAliveInterval=20',
        '-oStrictHostKeyChecking=no',
        `-oProxyCommand="'${ proxytunnel_cmd }' -p ${ boxname }:80 -d protonet:22 -o ${ boxname }"`,
        `${ username }@${ boxname }`
      ];
  }

  winston.info('[tunnel] Trying to establish SSH tunnel for syncthing...', {
    command: `${ ssh_cmd } ${ args.join(' ') }`
  });

  return new Promise(function(resolve, reject) {
    tunnel = exec(`"${ ssh_cmd }" ${ args.join(' ') }`);
    tunnelStarted = true;
    winston.info('[tunnel] Started.');

    tunnel.stdout.on('data', logBuffer('debug', '[tunnel]'));
    tunnel.stderr.on('data', logBuffer('warn', '[tunnel]'));

    tunnel.on('close', function(code) {
      tunnelStarted = false;
      winston.info('[tunnel] Stopped.', { code: code });
      if (code) reject(code);
      else resolve();
    });
  });
}

function sshTunnel(boxname, username) {
  function start() { return startTunnel(boxname, username); }

  helper.stat(ssh_key).then(function() {
    winston.debug('[tunnel] SSH key already exists');
    return start();
  }, function() {
    generateSSHKey().then(function() {
      return start();
    });
  });
}

function sshTunnelKill() {
  if (tunnel) {

    winston.info('[tunnel] Going to die...');

    if (isWin) {
      var taskkill = exec('taskkill /IM proxytunnel.exe /T /F');

      taskkill.stdout.on('data', logBuffer('debug', '[tunnel]'));
      taskkill.stderr.on('data', logBuffer('warn', '[tunnel]'));

      taskkill.on('close', function(error) {
        if (error) {
          winston.error("[taskkill] It's a Highlander!", { error: error });
        } else {
          tunnelStarted = false;
        }
      });
    } else if (tunnel) {
      tunnel.kill();
      tunnelStarted = false;
    }
  }
}

Native.getPublicKey            = getPublicKey;
Native.add_localhost_to_config = add_localhost_to_config;
Native.initSyncthingConfig     = initSyncthingConfig;
Native.sshTunnel               = sshTunnel;
Native.sshTunnelKill           = sshTunnelKill;
Native.syncthingClient         = syncthingClient;
Native.syncthing_cli           = syncthing_cli;
Native.syncthingCli            = syncthingCli;
Native.syncthingHome           = syncthingHome;

Native.registerDropHandle      = function(tray) { // jshint ignore:line
  // see darwin implementation
};

Native.downloadURL = 'http://packages2.protonet.info/protosync/current/';

const imagePath = path.join(app.getAppPath(), 'gfx');
Native.getTrayImagePath = function(status) {
  if (status === undefined) status = 0;
  return path.join(imagePath, `favicon${status}.png`);
};

const platformPath = path.join(__dirname, process.platform);
winston.debug(`[${ moduleName }] Specializing...`, { platform: process.platform });
module.exports = require(platformPath).apply(Native);
